/*
   FALCON - The Falcon Programming Language.
   FILE: vmstack.h

   Virtual machine stack and stack frame.
   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: Wed, 26 Mar 2008 20:09:53 +0100
   Last modified because:

   -------------------------------------------------------------------
   (C) Copyright 2008: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
   In order to use this file in its compiled form, this source or
   part of it you have to read, understand and accept the conditions
   that are stated in the LICENSE file that comes boundled with this
   package.
*/

/** \file
   Virtual machine stack and stack frame.
*/

#ifndef FALCON_VMSTACK_H
#define FALCON_VMSTACK_H

#include <falcon/setup.h>
#include <falcon/types.h>
#include <falcon/genericvector.h>
#include <falcon/string.h>
#include <falcon/basealloc.h>
#include <falcon/item.h>

namespace Falcon {

class VMStack;
class VMachine;
class Module;
class ItemVector;
class Symbol;

class FALCON_DYN_CLASS VMFrame: public BaseAlloc
{
public:
   enum { ALLOC_BLOCK = 16 };

   Symbol *m_symbol;
   const Module *m_module;
   ItemVector *m_globals;

   uint32 m_ret_pc;
   uint32 m_call_pc;
   bool m_initFrame;
   bool m_break;
   bool m_suspend;
   VMFrame *m_tryFrame;
   uint32 m_tryPos;
   Item m_sender;

   /** Number of parameters for this call frame. */
   uint32 m_param_count;

   /** Params are actually a pointer in the locals vector of the previous frame. */
   Item *m_params;

   /** Count of local variables.
      m_top can never be less than m_locals_count, or we'll have stack underflow
   */
   uint32 m_locals_count;

   /** Allocated size of the stack */
   uint32 m_allocated;

   /** Topmost element of the stack */
   uint32 m_top;

   /** Local variables + stack for this frame. */
   Item *m_locals;

   /** End frame handler function */
   ext_func_frame_t m_endFrameFunc;

   /** Previous stack frame (0 for topmost) */
   VMFrame *m_previous;

   /** Creates an empty stack frame. */
   VMFrame();

   /** Creates the stack frame from the current VM.
      This copies the live values in the VM and prepares the frame to execute the jump.
      Also, add this frame in the stack of the vm, if it exists.

      VM Stack initialization sequence:
       # Create an empty stack in a VMContext.
       # Apply it to the VM
       # Create a VMFrame with this constructor.
   */
   VMFrame( VMachine *vm, uint32 locals, uint32 paramCount );
   void configure( VMachine *vm, uint32 locals, uint32 paramCount );

   /** Destroys a frame. */
   ~VMFrame();

   void push( const Item &item );
   bool pop( Item &item );

   /** Returns the nth element in the stack */
   Item &local( uint32 pos ) { return m_locals[pos]; }
   const Item &local( uint32 pos ) const { return m_locals[pos]; }

   /** Returns the nth element in the stack */
   Item &param( uint32 pos ) { return m_params[pos]; }
   const Item &param( uint32 pos ) const { return m_params[pos]; }

   /** Returns the nth element from top (0 being the topmost vector element) */
   Item &back( uint32 pos ) { return m_locals[m_top - pos - 1]; }
   const Item &back( uint32 pos ) const { return m_locals[m_top - pos - 1]; }

   Item &topItem() {return m_locals[m_top - 1]; }
   const Item &topItem() const {return m_locals[m_top - 1]; }

   /** Ensure the local stack holds at least the given count of items.
      The number will be always round up to the nearest ALLOC_BLOCK multiple.
   */
   void reserve( uint32 count );

   /** Moves the stack pointer to the given position.
      If a reallocation is needed, the stack is espanded. If the size
      is less than half the allocated size, the stack allocation is resized down.
   */
   void resize( uint32 count );

   uint32 size() const { return m_top; }

   uint32 paramCount() const { return m_param_count; }
   VMFrame *prev() const { return m_previous; }
};

/** Falcon Virtual Machine Stack.
   A stack is a list of VMFrame items; each frame represents the local execution
   context of a function.
*/
class FALCON_DYN_CLASS VMStack: public BaseAlloc
{
public:
   VMFrame *m_bottomFrame;
   VMFrame *m_topFrame;

   /** Creates an empty stack.
      An empty stack is not very useful though. Add immediately a frame to it.
   */
   VMStack();

   /** Destroys the stack and all the stored frame. */
   ~VMStack();

   /** Appends a frame to the end of the frame list. */
   void addFrame( VMFrame *frame );

   /** Removes and returns the topmost frame */
   VMFrame *removeFrame();

   /** Returns the top frame. */
   VMFrame *topFrame() const { return m_topFrame; }

   /** Clears the stack, leaving only the topmost frame. */
   void clear();
};

}

#endif

/* end of vmstack.h */
